/********************************************

    chttp    -   http  library  for
     cserver
    David Marquez Hernandez  - 2016
    TFG UOC - Redes de Computadores
    Profesor  - Jesús Arribi Vilela

*********************************************/


#include "chttp.h"

/*
    NAME:   get_file_content_type
    DESC:   Identifies HTML content type
    INPUT:  pointer to filename and to  string to be filled
    OUTPUT: 0 an the string is filled with the HTML content type.
    NOTES:
*/

int get_file_content_type(char *filename,char *content_type){

    if (strstr(filename,".htm")!=NULL){
        strcpy(content_type,"text/html");
        return 0;
    } else if (strstr(filename,".txt")!=NULL){
        strcpy(content_type,"text/plain");
        return 0;
    } else if ((strstr(filename,".jpg")!=NULL) || (strstr(filename,".jpeg")!=NULL)){
        strcpy(content_type,"image/jpeg");
        return 0;
    } else if (strstr(filename,".png")!=NULL){
        strcpy(content_type,"image/png");
        return 0;
    } else if (strstr(filename,".gif")!=NULL){
        strcpy(content_type,"image/gif");
        return 0;
    };

    return 0;
}


/*
    NAME:   send_to_socket
    DESC:   Sends to string to already opened socket.
    INPUT:  socket to send the string to, pointer with the string to be sent.
    OUTPUT: 0 or errno.
    NOTES:
*/

int send_to_socket(int client_socket,char *string_to_send){
    int rw_result;

    if (logging==2) printf("Going to send %s with length %d to socket %d\n",string_to_send,(int)strlen(string_to_send),client_socket);
    //if ((rw_result = write(client_socket,string_to_send,strlen(string_to_send)))<0){
    if ((rw_result = send(client_socket,string_to_send,strlen(string_to_send),MSG_NOSIGNAL))<0){
        if ((errno==EBADF) || (errno==EPIPE)) {
            return 0; //The other side already closed the socket.
        } else {
            cerror_show(); //Shows ERRNO error in both screen and logfile.
            clogger_stop();
            close(client_socket);
            return errno;
        };
    };

    return 0;
}


/*
    NAME:   reply_http_request
    DESC:   Replys to a previous HTTP request, according to the that in the http headers struct
    INPUT:  client socket, pointer to http request struct, pointer to config, pointer to client ip string.
    OUTPUT: 0 or error number.
    NOTES:
*/

int reply_http_request(int client_socket, request_http_headers *headers,cconfig *config,  char *client_ip){
    int file_descriptor;
    char *response_line;
    char *file_content_type;
    struct stat file_stat;
    char *file_content;
    char  *log_line;
    char datetime[32];
    time_t now;
    struct tm time_s;

    response_line=malloc(256);
    if (response_line==NULL){
        cerror_show();
        return errno;
    }
    file_content_type=malloc(64);
    if (file_content_type==NULL){
        cerror_show();
        free(response_line);
        return errno;
    }
    log_line=malloc(MAX_LOG_LINE_SIZE);
    if (log_line==NULL){
        cerror_show();
        free(response_line);
        free(file_content_type);
        return errno;
    }
    if (strcmp(headers->method,"GET")==0){
        if ((file_descriptor = open(strcat(config->www_root,headers->resource),O_RDONLY,0))<0) {
            if (logging==2) printf("\n404\n");
            send_to_socket(client_socket,"HTTP/1.0 404 NOT FOUND\r\n");
            send_to_socket(client_socket,"Server: CSERVER\r\n");
            sprintf(response_line,"Content-Length: %d\r\n",(int)strlen(PAGE404));
            send_to_socket(client_socket,response_line);
            send_to_socket(client_socket,"Content-Type: text/html\r\n");
            send_to_socket(client_socket,"\r\n");
            send_to_socket(client_socket,PAGE404);
            if (logging>0) { //Only bother building the string if logging is activated, although we check again inside clogger.
                now=time(NULL);
                time_s=*(localtime_r(&now,&time_s));  //localtime_r is the thread-safe version of localtime
                strftime(datetime,32,"%d/%b/%Y:%T %z",&time_s);
                sprintf(log_line,"%s - [%s] \"%s %s HTTP/%.1f\" 404 %d\n",client_ip,datetime,headers->method,headers->resource,headers->version,(int)strlen(PAGE404));
                clogger_log(log_line);
            }

        } else { //200OK
            if (logging==2) printf("\n200\n");
            send_to_socket(client_socket,"HTTP/1.0 200 OK\r\n");
            send_to_socket(client_socket,"Server: CSERVER\r\n");

            if (fstat(file_descriptor, &file_stat) <0) {
                cerror_show(); //Shows ERRNO error in both screen and logfile.
                clogger_stop();
                close(client_socket);
                close(file_descriptor);
                free(response_line);
                free(file_content_type);
                free(log_line);
                return errno;
            }
            sprintf(response_line,"Content-Length: %d\r\n",(int)file_stat.st_size);
            send_to_socket(client_socket,response_line);
            get_file_content_type(headers->resource,(char *) file_content_type);
            sprintf(response_line,"Content-Type: %s\r\n",file_content_type);
            send_to_socket(client_socket,response_line);
            send_to_socket(client_socket,"\r\n");
            file_content=malloc(file_stat.st_size);
            if (file_content==NULL){
                cerror_show(); //Shows ERRNO error in both screen and logfile.
                clogger_stop();
                close(client_socket);
                close(file_descriptor);
                free(response_line);
                free(file_content_type);
                free(log_line);
                return errno;
            }

            if (read(file_descriptor,file_content,file_stat.st_size)<0) {
                cerror_show(); //Shows ERRNO error in both screen and logfile.
                clogger_stop();
                close(client_socket);
                free(file_content);
                close(file_descriptor);
                free(response_line);
                free(file_content_type);
                free(log_line);
                return errno;
            }
            //if (write(client_socket,file_content,file_stat.st_size)<0) {
            if (send(client_socket,file_content,file_stat.st_size,MSG_NOSIGNAL)<0) {
                if ((errno==EBADF) || (errno==EPIPE)) {
                    close(client_socket);
                    free(file_content);
                    close(file_descriptor);
                    free(response_line);
                    free(file_content_type);
                    free(log_line);
                    return 0; //The other side already closed the socket.
                } else {
                    cerror_show(); //Shows ERRNO error in both screen and logfile.
                    clogger_stop();
                    close(client_socket);
                    free(file_content);
                    close(file_descriptor);
                    free(response_line);
                    free(file_content_type);
                    free(log_line);
                    return errno;
                };
            }
            free(file_content);
            /* Apache Log Format
            127.0.0.1 - authenticatedUser [11/Oct/2002:17:51:26 -0700] "GET /apache_pb.gif HTTP/1.0" 200 2316*/
            if (logging>0) { //Only bother building the string if logging is activated, although we check again inside clogger.
                now=time(NULL);
                time_s=*(localtime_r(&now,&time_s));
                strftime(datetime,32,"%d/%b/%Y:%T %z",&time_s);
                sprintf(log_line,"%s - [%s] \"%s %s HTTP/%.1f\" 200 %d\n",client_ip,datetime,headers->method,headers->resource,headers->version,(int)file_stat.st_size);
                clogger_log(log_line);
            }
        }
    } else {
        if (logging==2) printf("NOT IMPLEMENTED\n");
        send_to_socket(client_socket,"HTTP/1.0 501 NOT IMPLEMENTED\r\n");
        send_to_socket(client_socket,"Server: CSERVER\r\n");
        sprintf(response_line,"Content-Length: %d\r\n",(int)strlen(PAGE501));
        send_to_socket(client_socket,response_line);
        send_to_socket(client_socket,"Content-Type: text/html\r\n");
        send_to_socket(client_socket,"\r\n");
        send_to_socket(client_socket,PAGE501);
        if (logging>0) { //Only bother building the string if logging is activated, although we check again inside clogger.
            now=time(NULL);
            time_s=*(localtime_r(&now,&time_s));
            strftime(datetime,32,"%d/%b/%Y:%T %z",&time_s);
            sprintf(log_line,"%s - [%s] \"%s %s HTTP/%.1f\" 501 %d\n",client_ip,datetime,headers->method,headers->resource,headers->version,(int)strlen(PAGE501));
            clogger_log(log_line);
        }

    }

    close(file_descriptor);
    free(log_line);
    free(response_line);
    free(file_content_type);
    return 0;
};



/*
    NAME:   read_http_request
    DESC:   Reads HTTP request and fills http header struct.
    INPUT:  client socket and pointer to http header struct to be filled.
    OUTPUT: 0 or error number.
    NOTES:  There is no maximum standard set for HTTP headers, although Apache
    uses 8k and MS IIS uses 16k. Therefore this function will use 8k.
    Max length per line supported is 1024.
*/

int read_http_request(int client_socket, request_http_headers *headers){

    char *buffer;
    char *line;
    int rw_result;
    int i,j;
    int header_line_number=0;
    char *internal_line_index;
    char *internal_line_index2;
    //char log_line[MAX_LOG_LINE_SIZE];
    char *log_line;



     log_line=malloc(MAX_LOG_LINE_SIZE);
    if (log_line==NULL){
        cerror_show();
        return 0;
    }

    buffer=malloc(MAX_HEADER_SIZE);
    if (buffer==NULL){
        cerror_show(); //Shows ERRNO error in both screen and logfile.
        clogger_stop();
        close(client_socket);
        return errno;
    }
    line=malloc(MAX_HEADER_LINE_SIZE);
    if (line==NULL){
        cerror_show(); //Shows ERRNO error in both screen and logfile.
        clogger_stop();
        close(client_socket);
        free(buffer);
        return errno;
    }

    memset(headers->method,0,MAX_HEADER_LINE_SIZE);
    memset(headers->resource,0,MAX_HEADER_LINE_SIZE);
    headers->version=0.0;
    memset(headers->host,0,MAX_HEADER_LINE_SIZE);
    memset(headers->user_agent,0,MAX_HEADER_LINE_SIZE);
    memset(headers->accept,0,MAX_HEADER_LINE_SIZE);
    memset(headers->accept_language,0,MAX_HEADER_LINE_SIZE);
    memset(headers->accept_encoding,0,MAX_HEADER_LINE_SIZE);
    memset(headers->connection,0,MAX_HEADER_LINE_SIZE);
    memset(headers->cache_control,0,MAX_HEADER_LINE_SIZE);
    memset(line,0,MAX_HEADER_LINE_SIZE);
    memset(buffer,0,MAX_HEADER_SIZE);

    if (logging==2) printf ("CSERVER reading request\n");
    if ((rw_result = read(client_socket,buffer,8192))<0) {
        cerror_show(); //Shows ERRNO error in both screen and logfile.
        clogger_stop();
        close(client_socket);
        free(buffer);
        free(line);
        return errno;
    }
    if (logging==2) printf ("CSERVER we have read %d characters\n",rw_result);
    if (logging==2) printf("Here is the message with length %d:\n %s\n",rw_result,buffer);

    //Process each message line:
    j=0;
    for (i=0;i<rw_result;i++){
        if (buffer[i]!='\n') {
            line[j]=buffer[i];
            j++;
        } else {
            //Process line:
            if (line[0]=='\n'){ //Last line of HTTP request headers
                break;
            }
            if (header_line_number==0) {   //First line is the method
                   //Get HTTP version
                if ((internal_line_index=strstr(line,"HTTP/"))!=NULL){ //We have version information
                    headers->version=(float)atof(internal_line_index+5);
                }

                if ( !strncmp(line, "GET ", 4) ) {
                    strcpy(headers->method,"GET");
                    //GET Resource
                    if ((internal_line_index=strchr(line,'/'))!=NULL){ //locate the first slash
                       internal_line_index2=strstr(internal_line_index+1,"HTTP/"); //locate the second slash
                       internal_line_index2--;
                       *internal_line_index2='\0'; //End the string there.
                        if (internal_line_index[strlen(internal_line_index)-1]=='/') { //We need to append index.html
                            strcat(internal_line_index,"index.html");
                        }
                        strcpy(headers->resource,internal_line_index);
                    }
                } else {
                    strcpy(headers->method,"NOT IMPLEMENTED");
                }



            } else { //Lines 2+
                if ( !strncmp(line, "Host:", 5) ) {
                    strncpy(headers->host,line+5,strchr(line,'\r')-line-5);
                } else if ( !strncmp(line, "User-Agent:", 11 )) {
                    strncpy(headers->user_agent,line+11,strchr(line,'\r')-line-11);
                } else if ( !strncmp(line, "Accept:", 7 )) {
                    strncpy(headers->accept,line+7,strchr(line,'\r')-line-7);
                } else if ( !strncmp(line, "Accept-Language:", 16 )) {
                    strncpy(headers->accept_language,line+16,strchr(line,'\r')-line-16);
                } else if ( !strncmp(line, "Accept-Encoding:", 16 )) {
                    strncpy(headers->accept_encoding,line+16,strchr(line,'\r')-line-16);
                } else if ( !strncmp(line, "Connection:", 11 )) {
                    strncpy(headers->connection,line+11,strchr(line,'\r')-line-11);
                } else if ( !strncmp(line, "Cache-Control:", 14 )) {
                    strncpy(headers->cache_control,line+14,strchr(line,'\r')-line-14);
                };
            };
            header_line_number++;
            //Reset line index
            j=0;
        }

    }


    if (logging==2) printf("HTTP HEADERS> METHOD %s RES %s VERSION %f HOST %s UA %s\n",headers->method,headers->resource,headers->version,headers->host,headers->user_agent);
    if (logging==2) printf("Acc %s ACc-Lang %s acc-enc %s Conn %s cache %s\n",headers->accept,headers->accept_language,headers->accept_encoding,headers->connection,headers->cache_control);
    sprintf(log_line,"RECEIVED %s %s %s\n",headers->method,headers->resource,headers->user_agent);
    clogger_log(log_line);

    free(buffer);
    free(line);
    free(log_line);
    return 0;
}
